// ---------------------------------------------------------------------------
// - Instance.cpp                                                            -
// - afnix engine - afnix instance class implementation                      -
// ---------------------------------------------------------------------------
// - This program is free software;  you can redistribute it  and/or  modify -
// - it provided that this copyright notice is kept intact.                  -
// -                                                                         -
// - This program  is  distributed in  the hope  that it will be useful, but -
// - without  any  warranty;  without  even   the   implied    warranty   of -
// - merchantability or fitness for a particular purpose.  In no event shall -
// - the copyright holder be liable for any  direct, indirect, incidental or -
// - special damages arising in any way out of the use of this software.     -
// ---------------------------------------------------------------------------
// - copyright (c) 1999-2006 amaury darsch                                   -
// ---------------------------------------------------------------------------

#include "Cons.hpp"
#include "Vector.hpp"
#include "Method.hpp"
#include "Closure.hpp"
#include "Runnable.hpp"
#include "Instance.hpp"
#include "Exception.hpp"

namespace afnix {
  // the instance eval quarks
  static const long QUARK_THIS  = String::intern ("this");
  static const long QUARK_META  = String::intern ("meta");
  static const long QUARK_SETM  = String::intern ("set-meta");
  static const long QUARK_INIT  = String::intern ("initialize");
  static const long QUARK_SUPER = String::intern ("super");

  // create a default instance

  Instance::Instance (void) {
    // initialize data
    p_iset  = nilp;
    p_class = nilp;
    p_super = nilp;
    d_const = false;
  }

  // create a new instance with a class and some arguments

  Instance::Instance (Runnable* robj, Nameset* nset, Cons* args, Class* cls) {
    // initialize data
    p_iset  = nilp;
    p_class = nilp;
    p_super = nilp;
    d_const = false;
    // set the instance class
    setmeta (robj, nset, args, cls);
  }

  // destroy this instance

  Instance::~Instance (void) {
    Object::dref (p_iset);
    Object::dref (p_class);
    Object::dref (p_super);
  }

  // return the class name

  String Instance::repr (void) const {
    return "Instance";
  }

  // make this instance a shared object

  void Instance::mksho (void) {
    if (p_shared != nilp) return;
    Object::mksho ();
    if (p_class != nilp) p_class->mksho ();
    if (p_super != nilp) p_super->mksho ();
    if (p_iset  != nilp) p_iset->mksho  (); 
  }

  // set the instance meta class
  
  Object* Instance::setmeta (Runnable* robj, Nameset* nset, Cons* args) {
    wrlock ();
    // trivial check first
    if ((args == nilp) || (args->length () < 1)) {
      unlock ();
      throw Exception ("argument-error",
                       "missing or too many arguments with set-meta");
    }
    // try to get the class
    Object* car = args->getcar ();
    Object* obj = (car == nilp) ? nilp : car->eval (robj, nset);
    Class*  cls = dynamic_cast <Class*> (obj);
    if (cls == nilp) {
      unlock ();
      throw Exception ("type-error", "invalid object in set-meta");
    }
    // set the meta class
    try {
      Cons* cdr      = args->getcdr ();
      Object* result = setmeta (robj, nset, cdr, cls);
      unlock ();
      return result;
    } catch (...) {
      unlock ();
      throw;
    }
  }

  // set the instance meta class

  Object* Instance::setmeta (Runnable* robj, Nameset* nset, Cons* args, 
			     Class* cls) {
    wrlock ();
    // protect the class
    Object::iref (cls);
    // clean the everything locally
    Object::dref (p_iset);
    Object::dref (p_class);
    // create the instance local set and bind this
    Object::iref   (p_iset = new Localset);
    p_iset->symcst (QUARK_THIS, this);
    // save the meta class
    p_class = cls;
    // bind the default symbols
    try {
      const Qarray& mdata = cls->getmdata ();
      if (mdata.length () != 0) {
	long len = mdata.length ();
	for (long i = 0; i < len; i++)
	  p_iset->symdef (mdata.get (i), (Object*) nilp);
      }
    } catch (...) {
      unlock ();
      throw;
    }
    // eventually make everything shared
    if (p_shared != nilp) mksho ();
    // find the initial form
    Object* result = nilp;
    try {
      Object* iobj = p_class->find (QUARK_INIT);
      Object* form = (iobj == nilp) ? nilp : iobj->eval (robj, nset);
      if (form != nilp) {
	p_iset->setparent (nset);
	result = form->apply (robj, p_iset, args);
	p_iset->setparent ((Nameset*) nilp);
      } 
    } catch (...) {
      Object::iref   (this);
      p_iset->remove (QUARK_THIS);
      Object::tref   (this);
      unlock ();
      throw;
    }
    Object::iref   (this);
    p_iset->remove (QUARK_THIS);
    Object::tref   (this);
    robj->post (result);
    unlock ();
    return result;
  }

  // set the instance super value

  Object* Instance::setsuper (Object* object, const bool flag) {
    wrlock ();
    if (d_const == true) {
      unlock ();
      throw Exception ("const-error", "const violation with super member");
    }
    Object::iref (object);
    Object::dref (p_super);
    p_super = object;
    d_const = flag;
    unlock ();
    return object;
  }

  // create a new instance in a generic way

  Object* Instance::mknew (Vector* argv) {
    // get the number of arguments
    long argc = (argv == nilp) ? 0 : argv->length ();

    // check 0 argument
    if (argc == 0) return new Instance;
    // illegal arguments
    throw Exception ("argument-error", "too many arguments with instance");
  }

  // return true if the given quark is defined

  bool Instance::isquark (const long quark, const bool hflg) const {
    // check for local quarks
    if (quark == QUARK_THIS)  return true;
    if (quark == QUARK_META)  return true;
    if (quark == QUARK_SETM)  return true;
    if (quark == QUARK_INIT)  return true;
    if (quark == QUARK_SUPER) return true;
    // locak and check in the instance set
    rdlock ();
    if ((p_iset != nilp) && (p_iset->exists (quark) == true)) {
      unlock ();
      return true;
    }
    // check in the class
    if ((p_class != nilp) && (p_class->isquark (quark, hflg) == true)) {
      unlock ();
      return true;
    }
    // check in the super instance if hierarchical
    if (p_super != nilp) {
      bool result = hflg ? p_super->isquark (quark, hflg) : false;
      unlock ();
      return result;
    }
    // check in the instance base object
    bool result = hflg ? Object::isquark (quark, hflg) : false;
    unlock ();
    return result;
  }

  // create a const symbol in the instance local set

  Object* Instance::cdef (Runnable* robj, Nameset* nset, const long quark,
			  Object* object) {
    // check for the super quark
    if (quark == QUARK_SUPER) return setsuper (object, true);
    // check for reserved quark
    if (quark == QUARK_SETM) {
      throw Exception ("instance-error", "trying to set the meta object");
    }
    // create a local instance set if needed
    if (p_iset == nilp) {
      Object::iref   (p_iset = new Localset);
      p_iset->symcst (QUARK_THIS, this);
      mksho ();
    }
    // get the object from the localset
    return p_iset->cdef (robj, nset, quark, object);
  }

  // bind a symbol in the instance local set

  Object* Instance::vdef (Runnable* robj, Nameset* nset, const long quark,
			  Object* object) {
    // check for super
    if (quark == QUARK_SUPER) return setsuper (object, false);
    // check for reserved quark
    if (quark == QUARK_SETM) {
      throw Exception ("instance-error", "trying to set the meta object");
    }
    // create a local instance set if needed
    if (p_iset == nilp) {
      Object::iref   (p_iset = new Localset);
      p_iset->symcst (QUARK_THIS, this);
      mksho ();
    }
    // look in the instance local set
    if (p_iset != nilp) {
      Object* iobj = p_iset->find (quark);
      if (iobj != nilp) return iobj->vdef (robj, nset, object);
    }
    // look in the class
    if (p_class != nilp) {
      Object* cobj = p_class->find (quark);
      if (cobj != nilp) return cobj->vdef (robj, nset, object);
    }
    // bind locally
    if (p_iset != nilp) {
      return p_iset->vdef (robj, nset, quark, object);
    }
    throw Exception ("instance-error", "cannot access local instance set");
  }
  
  // evaluate an instance member

  Object* Instance::eval (Runnable* robj, Nameset* nset, const long quark) {
    // check for super 
    if (quark == QUARK_SUPER) return p_super;
    // check for meta and set-meta
    if (quark == QUARK_META) return p_class;
    if (quark == QUARK_SETM) return new Method (quark, this, false);
    // create a local instance set if needed
    if (p_iset == nilp) {
      Object::iref   (p_iset = new Localset);
      p_iset->symcst (QUARK_THIS, this);
      mksho ();
    }
    // check in the instance local set
    if (p_iset != nilp) {
      Object* obj = p_iset->find (quark);
      if (obj != nilp) {
	Object* result = obj->eval (robj, nset);
	if (dynamic_cast <Closure*> (result) == nilp) return result;
	return new Method (result, this);
      }
    }
    // check in the class
    if (p_class != nilp) {
      Object* obj = p_class->find (quark);
      if (obj != nilp) {
	Object* result = obj->eval (robj, nset);
	if (dynamic_cast <Closure*> (result) == nilp) return result;
	return new Method (result, this);
      }
    }
    // check in the super instance
    if (p_super != nilp) {
      Object* result = p_super->eval (robj, nset, quark);
      if (dynamic_cast <Closure*> (result) == nilp) return result;
      return new Method (result, this);
    }
    // last resort is a method
    return new Method (quark, this, true);
  }

  // evaluate an object with a set of arguments by quark

  Object* Instance::apply (Runnable* robj, Nameset* nset, const long quark,
			   Cons* args) {
    // check for set-meta quark
    if (quark == QUARK_SETM) {
      Object* result = setmeta (robj, nset, args);
      return result;
    }
    // evaluate and apply
    Object* obj = eval (robj, nset, quark);
    return apply (robj, nset, obj, args);
  }
    
  // evaluate an object within this instance

  Object* Instance::apply (Runnable* robj, Nameset* nset, Object* object,
			   Cons* args) {
    // basic reject - as usual
    if (object == nilp) return nilp;
    // create a local instance set if needed
    if (p_iset == nilp) {
      Object::iref   (p_iset = new Localset);
      p_iset->symcst (QUARK_THIS, this);
      mksho ();
    }
    // rebind the local set before the call
    Localset* lset = new Localset (p_iset);
    lset->setparent (nset);
    lset->symcst    (QUARK_THIS, this);
    Object* result = nilp;
    // let's make the call
    try {
      result = object->apply (robj, lset, args);
      lset->reset ();
      delete lset;
    } catch (...) {
      lset->reset ();
      delete lset;
      throw;
    }
    return result; 
  }
}
