// ---------------------------------------------------------------------------
// - Stream.cpp                                                              -
// - standard object library - base stream class implementation              -
// ---------------------------------------------------------------------------
// - This program is free software;  you can redistribute it  and/or  modify -
// - it provided that this copyright notice is kept intact.                  -
// -                                                                         -
// - This program  is  distributed in  the hope  that it will be useful, but -
// - without  any  warranty;  without  even   the   implied    warranty   of -
// - merchantability or fitness for a particular purpose.  In no event shall -
// - the copyright holder be liable for any  direct, indirect, incidental or -
// - special damages arising in any way out of the use of this software.     -
// ---------------------------------------------------------------------------
// - copyright (c) 1999-2006 amaury darsch                                   -
// ---------------------------------------------------------------------------

#include "Item.hpp"
#include "Stream.hpp"
#include "Vector.hpp"
#include "Exception.hpp"
#include "cloc.hpp"

namespace afnix {

  // -------------------------------------------------------------------------
  // - quark section                                                         -
  // -------------------------------------------------------------------------

  // the stream eval quarks
  static const long QUARK_BYTE    = String::intern ("BYTE");
  static const long QUARK_UTF8    = String::intern ("UTF-8");
  static const long QUARK_STREAM  = String::intern ("Stream");

  // the stream supported quarks
  static const long QUARK_GETSMOD = String::intern ("get-coding-mode");
  static const long QUARK_SETSMOD = String::intern ("set-coding-mode");

  // the quark table info
  static const long QUARK_TABLE_LEN = 2;
  static const long QUARK_TABLE_MAP [QUARK_TABLE_LEN] = {
    QUARK_GETSMOD,
    QUARK_SETSMOD
  };

  // -------------------------------------------------------------------------
  // - private functions section                                             -
  // -------------------------------------------------------------------------

  // map an enumeration item to a stream codingmode
  static inline Stream::t_smod item_to_smod (const Item& item) {
    // check for a stream item 
    if (item.gettid () != QUARK_STREAM)
      throw Exception ("item-error", "item is not a stream item");
    // map the item to the enumeration
    long quark = item.getquark (); 
    if (quark == QUARK_BYTE) return Stream::BYTE;
    if (quark == QUARK_UTF8) return Stream::UTF8;
    throw Exception ("item-error", "cannot map item to stream coding mode");
  }

  // map a coding mode to an item
  static inline Item* smod_to_item (const Stream::t_smod smod) {
    if (smod == Stream::BYTE) return new Item (QUARK_STREAM, QUARK_BYTE);
    if (smod == Stream::UTF8) return new Item (QUARK_STREAM, QUARK_UTF8);
    throw Exception ("mode-error", "cannot map stream coding mode to item");
  }

  // -------------------------------------------------------------------------
  // - public static section                                                 -
  // -------------------------------------------------------------------------

  Stream::t_smod Stream::getssm (void) {
    // get the system codeset
    t_codeset codeset = c_getcset ();
    // map the code set
    switch (codeset) {
    case LOC_ASCII:
    case LOC_LATIN:
      return BYTE;
      break;
    case LOC_UTF8:
      return UTF8;
      break;
    default:
      break;
    }
    return BYTE;
  }

  // -------------------------------------------------------------------------
  // - stream class section                                                  -
  // -------------------------------------------------------------------------

  // create a default stream

  Stream::Stream (void) {
    d_smod = BYTE;
  }

  // create a stream by coding mode

  Stream::Stream (const t_smod smod) {
    d_smod = smod;
  }

  // return the stream coding mode

  Stream::t_smod Stream::getsmod (void) const {
    rdlock ();
    t_smod result = d_smod;
    unlock ();
    return result;
  }

  // set the stream mode

  void Stream::setsmod (const t_smod smod) {
    wrlock ();
    d_smod = smod;
    unlock ();
  }

  // evaluate a quark statically

  Object* Stream::meval (Runnable* robj, Nameset* nset, const long quark) {
    if (quark == QUARK_BYTE) return new Item (QUARK_STREAM, QUARK_BYTE);
    if (quark == QUARK_UTF8) return new Item (QUARK_STREAM, QUARK_UTF8);
    throw Exception ("eval-error", "cannot evaluate member",
                     String::qmap (quark));
  }

  // return true if the given quark is defined

  bool Stream::isquark (const long quark, const bool hflg) const {
    rdlock ();
    for (long i = 0; i < QUARK_TABLE_LEN; i++) {
      if (quark == QUARK_TABLE_MAP[i]) {
	unlock ();
	return true;
      }      
    }
    bool result = hflg ? Object::isquark (quark, hflg) : false;
    unlock ();
    return result;
  }

  // apply this stream class with a set of arguments and a quark

  Object* Stream::apply (Runnable* robj, Nameset* nset, const long quark,
			 Vector* argv) {
    // get the number of arguments
    long argc = (argv == nilp) ? 0 : argv->length ();
    
    // dispatch 0 argument
    if (argc == 0) {
      if (quark == QUARK_GETSMOD) return smod_to_item (getsmod ());
    }

    // dispatch one argument
    if (argc == 1) {
      if (quark == QUARK_SETSMOD) {
	Object* iobj = argv->get (0);
	Item*   item = dynamic_cast <Item*> (iobj);
	if (item == nilp) {	  
	  throw Exception ("argument-error", 
			   "invalid arguments with set-mode");
	}
	t_smod smod = item_to_smod (*item);
	setsmod (smod);
	return nilp;
      }
    }
    // apply these arguments with the object
    return Object::apply (robj, nset, quark, argv);
  }
}
