// ---------------------------------------------------------------------------
// - Terminal.cpp                                                            -
// - afnix standard library - terminal class implementation                  -
// ---------------------------------------------------------------------------
// - This program is free software;  you can redistribute it  and/or  modify -
// - it provided that this copyright notice is kept intact.                  -
// -                                                                         -
// - This program  is  distributed in  the hope  that it will be useful, but -
// - without  any  warranty;  without  even   the   implied    warranty   of -
// - merchantability or fitness for a particular purpose.  In no event shall -
// - the copyright holder be liable for any  direct, indirect, incidental or -
// - special damages arising in any way out of the use of this software.     -
// ---------------------------------------------------------------------------
// - copyright (c) 1999-2006 amaury darsch                                   -
// ---------------------------------------------------------------------------

#include "Vector.hpp"
#include "Boolean.hpp"
#include "Terminal.hpp"
#include "Character.hpp"
#include "Exception.hpp"
#include "ctrm.hpp"

namespace afnix {

  // the terminal supported quarks
  static const long QUARK_GETPP     = String::intern ("get-primary-prompt");
  static const long QUARK_GETSP     = String::intern ("get-secondary-prompt");
  static const long QUARK_SETPP     = String::intern ("set-primary-prompt");
  static const long QUARK_SETSP     = String::intern ("set-secondary-prompt");
  static const long QUARK_READLINE  = String::intern ("read-line");

  // the quark table info
  static const long QUARK_TABLE_LEN = 5;
  static const long QUARK_TABLE_MAP [QUARK_TABLE_LEN] = {
    QUARK_GETPP,
    QUARK_GETSP,
    QUARK_SETPP,
    QUARK_SETSP,
    QUARK_READLINE
  };

  // default prompts
  static const char* DEFAULT_PROMPT1 = "(axi) ";
  static const char* DEFAULT_PROMPT2 = "(...) ";
  
  // create a default terminal

  Terminal::Terminal (void) {
    // prompt initialize
    d_prt1 = DEFAULT_PROMPT1;
    d_prt2 = DEFAULT_PROMPT2;
    // save the input stream state
    save ();
  }

  // destroy this terminal

  Terminal::~Terminal (void) {
    // restore terminal state
    restore ();
  }

  // return the class name

  String Terminal::repr (void) const {
    return "Terminal";
  }

  // kill all editing in the current line

  void Terminal::kill (void) {
    long num = d_lbuf.setce ();
    mover (num);
    num = d_lbuf.length ();
    erasel (num);
    d_lbuf.reset ();
  }

  // replace an editing line with another one

  void Terminal::replace (const String& line) {
    kill ();
    d_lbuf.add (line);
    Output::write (line);
  }
    
  // read a line and return a string

  String Terminal::readline (bool pflag) {
    // reset the readline buffer
    d_lbuf.reset ();
    
    // check for the eof
    if (d_eof == true) return eofc;
    
    // turn off canonical mode and reset error mode
    bool canon = nocanon ();
    temode (false);

    // ok, print the prompt
    pflag ? Output::write (d_prt1) : Output::write (d_prt2);
    
    // loop until we have an eol/eof
    while (1) {
      t_quad w = wread ();

      // check for backspace
      if (isbs (w) == true) {
	bool status = d_lbuf.erase ();
	if (status == true) erasel (1);
	continue;
      }
      
      // check for delete
      if (isdel (w) == true) {
	bool status = d_lbuf.del ();
	if (status == true) del ();
	continue;
      }
      
      // check for insert
      if (isins (w) == true) {
	d_insert = !d_insert;
	d_lbuf.setimode (d_insert);
	continue;
      }
      
      // check for move to begin line
      if (w == WIDE_STDBL) {
	long num = d_lbuf.setcb ();
	movel (num);
	continue;
      }
      
      // check for move to end of line
      if (w == WIDE_STDEL) {
	long num = d_lbuf.setce ();
	mover (num);
	continue;
      }
      
      // check for kill everything
      if (w == WIDE_STDKB) {
	kill ();
	continue;
      }
      
      // check for kill to eol
      if (w == WIDE_STDKL) {
	long num = d_lbuf.setce ();
	mover (num);
	d_lbuf.kill (num);
	erasel (num);
	continue;
      }

      // check for refresh
      if (w == WIDE_STDRF) {
	String rline = pflag ? d_prt1 : d_prt2;
        rline = rline + d_lbuf.tostring ();
        bool status = refresh (rline);
        if (status == true) d_lbuf.setce ();
        continue;
      }

      // check for move to left
      if (isle (w) == true) {
	bool status = d_lbuf.movel ();
	if (status == true) movel  (1);
	continue;
      }
      
      // check for move to right
      if (isri (w) == true) {
	bool status = d_lbuf.mover ();
	if (status == true) mover  (1);
	continue;
      }

      // check for a move down
      if (isdo (w) == true) {
	if (d_cilo.istop () == true) {
	  kill ();
	  continue;
	}
	String* data = dynamic_cast <String*> (d_cilo.getup ());
	if (data == nilp)
	  kill ();
	else
	  replace (*data);
	continue;
      }
      
      // check for a move up
      if (isup (w) == true) {
	if (d_cilo.isbottom () == true) continue;
	String* data = dynamic_cast <String*> (d_cilo.getdown ());
	if (data == nilp)
	  kill ();
	else
	  replace (*data);
	continue;
      }

      // make sure the cursor is reset to the end for a eol/eof
      if (w == WIDE_STDNL) d_lbuf.setce ();
      if (w == WIDE_STDEF) d_lbuf.setce ();
      
      // check for printable - if not we reject - we accept eol/eof as well
      if (isprintable (w) == false) continue;

      // process normal character
      if ((w == WIDE_STDNL) || (w == WIDE_STDEF)) {
	if (d_lbuf.length () > 0) 
	  d_cilo.add (new String (d_lbuf.tostring()));
	d_lbuf.add (w);
	write (eolc);
	break;
      }
      d_lbuf.add (w);
      if (canon == true) insert (w);
    }
    // this is it - return the string buffer
    restore ();
    return d_lbuf.tostring ();
  }

  // set the primary prompt

  void Terminal::setpp (const String& value) {
    wrlock ();
    d_prt1 = value;
    unlock ();
  }

  // set the seconday prompt

  void Terminal::setsp (const String& value) {
    wrlock ();
    d_prt2 = value;
    unlock ();
  }

  // get the primary prompt

  String Terminal::getpp (void) const {
    rdlock ();
    String result = d_prt1;
    unlock ();
    return result;
  }

  // get the secondary prompt

  String Terminal::getsp (void) const {
    rdlock ();
    String result = d_prt2;
    unlock ();
    return result;
  }

  // create a new terminal in a generic way

  Object* Terminal::mknew (Vector* argv) {
    long argc = (argv == nilp) ? 0 : argv->length ();
    if (argc != 0) 
      throw Exception ("argument-error", "invalid arguments with terminal"); 
    return new Terminal;
  }

  // return true if the given quark is defined

  bool Terminal::isquark (const long quark, const bool hflg) const {
    rdlock ();
    for (long i = 0; i < QUARK_TABLE_LEN; i++) {
      if (quark == QUARK_TABLE_MAP[i]) {
	unlock ();
	return true;
      }      
    }
    bool result = hflg ? InputTerm::isquark (quark, hflg) : false;
    if (result == false) {
      result = hflg ? OutputTerm::isquark (quark, hflg) : false;
    }
    unlock ();
    return result;
  }

  // apply this terminal with a set of arguments and a quark

  Object* Terminal::apply (Runnable* robj, Nameset* nset, const long quark,
			   Vector* argv) {
    // check input term quark
    if (InputTerm::isquark (quark, true) == true)
      return InputTerm::apply(robj, nset, quark, argv);
    // check output term quark
    if (OutputTerm::isquark (quark, true) == true)
      return OutputTerm::apply(robj, nset, quark, argv);

    // get the number of arguments
    long argc = (argv == nilp) ? 0 : argv->length ();

    // dispatch 0 argument
    if (argc == 0) {
      if (quark == QUARK_READLINE) return new String    (readline (true));
      if (quark == QUARK_GETPP)    return new String    (getpp ());
      if (quark == QUARK_GETSP)    return new String    (getsp ());
    }

    // dispatch one argument
    if (argc == 1) {
      if (quark == QUARK_READLINE) {
	bool flag = argv->getbool (0);
	String* result = new String (readline (flag));
	return result;
      }
      if (quark == QUARK_SETPP) {
	String val = argv->getstring (0);
	setpp (val);
	return nilp;
      }
      if (quark == QUARK_SETSP) {
	String val = argv->getstring (0);
	setsp (val);
	return nilp;
      }
    }
    
    // call the object method
    return Object::apply (robj, nset, quark, argv);
  }
}
